#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/syscall.h>

#define MB (1024 * 1024)

// 获取线程ID
pid_t gettid() {
    return syscall(SYS_gettid);
}

// 线程函数
void *thread_function(void *arg) {
    // 接收传递的参数
    int *args = (int *)arg;
    int file_size_mb = args[0]; // 文件大小（MB）
    int thread_num = args[1];   // 线程编号

    // 打印进程ID和线程ID
    printf("Process ID: %d, Thread ID: %d\n", getpid(), gettid());

    // 创建文件名
    char filename[20];
    sprintf(filename, "file_%d.txt", thread_num);

    // 打开文件
    FILE *file = fopen(filename, "w");
    if (file == NULL) {
        perror("Error opening file");
        pthread_exit(NULL);
    }

    // 写入文件
    for (int i = 0; i < file_size_mb * MB; ++i) {
        fputc('A', file);
    }

    // 关闭文件
    fclose(file);
    pthread_exit(NULL);
}

int main(int argc, char *argv[]) {
    if (argc != 3) {
        printf("Usage: %s <number of threads> <file size in MB>\n", argv[0]);
        return 1;
    }

    int num_threads = atoi(argv[1]);       // 获取线程数
    int file_size_mb = atoi(argv[2]);      // 获取文件大小（MB）

    pthread_t threads[num_threads];
    int thread_args[num_threads][2];

    // 创建线程
    for (int i = 0; i < num_threads; ++i) {
        thread_args[i][0] = file_size_mb;
        thread_args[i][1] = i + 1;
        if (pthread_create(&threads[i], NULL, thread_function, &thread_args[i]) != 0) {
            perror("Error creating thread");
            return 1;
        }
    }

    // 等待线程结束
    for (int i = 0; i < num_threads; ++i) {
        if (pthread_join(threads[i], NULL) != 0) {
            perror("Error joining thread");
            return 1;
        }
    }

    return 0;
}

