#include <uapi/linux/ptrace.h>
#include <linux/sched.h>
/* #include <linux/compaction.h> */

typedef struct {
	u32 id;
}mem_key_t;

typedef struct {
	char comm[TASK_COMM_LEN];	//线程名
	u64  alloc_size;
	u64  free_size;	
}mem_info_t;

BPF_HASH(mem_info, mem_key_t, mem_info_t);

TRACEPOINT_PROBE(kmem, mm_page_alloc) {
	mem_key_t    key;
	mem_info_t  *info, zero={};
	u64 len = PAGE_SIZE << (args->order);
	
	/* key.id = bpf_get_current_pid_tgid(); */
	/* key.id = bpf_get_current_pid_tgid() >> 32; */
	MACRO_GET_ID
	
	info = mem_info.lookup_or_try_init(&key, &zero);
	if (info) {
		bpf_get_current_comm(&info->comm, TASK_COMM_LEN);
		info->alloc_size += len;
		mem_info.update(&key, info);
	}

	return 0;
}

TRACEPOINT_PROBE(kmem, mm_page_free) {
 	mem_key_t    key;
	mem_info_t  *info, zero={};

	/* key.id = bpf_get_current_pid_tgid(); */
	/* key.id = bpf_get_current_pid_tgid() >> 32; */
	MACRO_GET_ID
	
	info = mem_info.lookup_or_try_init(&key, &zero);
	if (info) {
		bpf_get_current_comm(&info->comm, TASK_COMM_LEN);
		info->free_size  += (PAGE_SIZE << args->order);
		mem_info.update(&key, info);
	}

	return 0;
}
