#include <uapi/linux/ptrace.h>
#include <linux/sched.h>
#include <linux/blkdev.h>
#include <linux/blk-mq.h>
#include <linux/sched.h>
#include <linux/cred.h>
#include <linux/sched/prio.h>


typedef struct {
	u32 id;
}io_key_t;

typedef struct {
	u32 uid;
	s32 prio;
	u32 policy;
	u64 write_bytes;
	u64 read_bytes;
	u64 io;
	char comm[TASK_COMM_LEN];
}io_info_t;
BPF_HASH(io_info, io_key_t, io_info_t, 1024);

/* ssize_t vfs_write(struct file *file, const char __user *buf, size_t count, loff_t *pos) */
ssize_t vfs_write_hook(struct pt_regs *ctx, struct file *file, const char __user *buf, size_t count, loff_t *pos)
{
	io_key_t key={};
	io_info_t *info, zero={};

		
	/* key.id = bpf_get_current_pid_tgid(); */
	/* key.id = bpf_get_current_pid_tgid() >> 32; */
	MACRO_GET_ID
	info = io_info.lookup_or_try_init(&key, &zero);
	if (info) {
		struct task_struct *task;
		struct cred cred;
		task = (struct task_struct *)bpf_get_current_task();
		if (task) {
			info->uid = task->cred->uid.val;
			info->policy = task->policy;
			info->prio  = task->static_prio - DEFAULT_PRIO;
		}
		
		bpf_get_current_comm(&info->comm, TASK_COMM_LEN);
		info->io += 1;
		info->write_bytes += count;
		/* info->prio = 2; */

		/* current->cred->gid; */
		io_info.update(&key, info);
	}

	return  0;
}

/* ssize_t vfs_read_hook(struct file *file, char __user *buf, size_t count, loff_t *pos) */
ssize_t vfs_read_hook(struct pt_regs *ctx, struct file *file, char __user *buf, size_t count, loff_t *pos)
{
	io_key_t key={};
	io_info_t *info, zero={};
	
	/* key.id = bpf_get_current_pid_tgid(); */
	/* key.id = bpf_get_current_pid_tgid() >> 32; */
	MACRO_GET_ID

	info = io_info.lookup_or_try_init(&key, &zero);
	if (info) {
		struct task_struct *task;
		struct cred cred;
		task = (struct task_struct *)bpf_get_current_task();
		if (task) {
			info->uid = task->cred->uid.val;
			info->policy = task->policy;
			info->prio  = task->static_prio - DEFAULT_PRIO;
			
		}

		bpf_get_current_comm(&info->comm, TASK_COMM_LEN);
		info->io += 1;
		info->read_bytes += count;
		io_info.update(&key, info);
	}

	return  0;
}

/* **************************************************************** */
typedef  struct {
	dev_t dev;
	u32 pad;
	sector_t sector;
} disk_key_t;

typedef struct {
	u32 id;
	u32 uid;
	s32 prio;
	u32 policy;
	u64 write_bytes;
	u64 read_bytes;
	char comm[TASK_COMM_LEN];
} disk_io_who_t;
BPF_HASH(who_info, disk_key_t, disk_io_who_t, 1024);

struct tp_args {
    u64 __unused__;
    dev_t dev;
    sector_t sector;
    unsigned int nr_sector;
    unsigned int bytes;
    char rwbs[8];
    char comm[16];
    char cmd[];
};


static dev_t ddevt(struct gendisk *disk) {
    return (disk->major  << 20) | disk->first_minor;
}

/* void blk_account_io_start(struct request *rq) */
void blk_account_io_start_hook(struct pt_regs *ctx, struct request *rq) {
	disk_io_who_t *who, zero = {};
	disk_key_t key = {
		.dev = ddevt(rq->__RQ_DISK__),
			.sector = rq->__sector,
	};

	who = who_info.lookup_or_try_init(&key, &zero);
	if (who) {
		if (bpf_get_current_comm(&who->comm, sizeof(who->comm)) == 0) {
			struct task_struct *task;
			struct cred cred;

			task = (struct task_struct *)bpf_get_current_task();
			if (task) {
				who->uid = task->cred->uid.val;
				who->policy = task->policy;
				who->prio  = task->static_prio - DEFAULT_PRIO;

				if (rq->cmd_flags == 1) {
					who->write_bytes += rq->__data_len;
				} else {
					who->read_bytes += rq->__data_len;
				}
			}
			MACRO_GET_DISK_ID
			who_info.update(&key, who);
		}
	}
}

/* void blk_account_io_done(struct request *req, u64 now) */
void blk_account_io_done_hook(struct pt_regs *ctx, struct request *rq, u64 now) {
	io_key_t key = {};
	io_info_t *info, zero={};
	disk_io_who_t *who;
	disk_key_t d_k = {
		.dev = ddevt(rq->__RQ_DISK__),
		.sector = rq->__sector,
	};

	who = who_info.lookup(&d_k);
	if (who == 0) {
		return;
	}

	key.id = who->id;
	info = io_info.lookup_or_try_init(&key, &zero);
	if (info) {

		info->uid = who->uid;
		info->policy = who->policy;
		info->prio = who->prio;
		info->io += 1;
		info->read_bytes += who->read_bytes;
		info->write_bytes += who->write_bytes;
		__builtin_memcpy(info->comm, who->comm, TASK_COMM_LEN);
		io_info.update(&key, info);
	}

	who_info.delete(&d_k);

	return ;
}


int trace_pid_start_tp(struct tp_args *args)
{
	disk_io_who_t *who, zero = {};
	disk_key_t key = {
		.dev = args->dev,
		.sector = args->sector,
	};

    who = who_info.lookup_or_try_init(&key, &zero);
	if (who) {
		struct task_struct *task;
		struct cred cred;
		task = (struct task_struct *)bpf_get_current_task();
		if (task) {
			who->uid = task->cred->uid.val;
			who->policy = task->policy;
			who->prio  = task->static_prio - DEFAULT_PRIO;
		}

		bpf_get_current_comm(who->comm, sizeof(who->comm));
		MACRO_GET_DISK_ID
		who_info.update(&key, who);
	}

	return 0;
}

// time block I/O
int trace_req_start(struct pt_regs *ctx, struct request *req)
{
	disk_io_who_t *who, zero = {};
	disk_key_t key = {
		.dev = ddevt(req->__RQ_DISK__),
		.sector = req->__sector,
	};

	who = who_info.lookup_or_try_init(&key, &zero);
	if (who) {
        u32 rw_flag = !!((req->cmd_flags & REQ_OP_MASK) == REQ_OP_WRITE);
		if (rw_flag) {
			who->write_bytes += req->__data_len;
		} else {
			who->read_bytes += req->__data_len;
		}
		who_info.update(&key, who);
	}

    return 0;
}

int trace_req_completion_tp(struct tp_args *args)
{
	io_key_t key = {};
	io_info_t *info, zero={};	
	disk_io_who_t *who;
	disk_key_t d_k = {
        .dev = args->dev,
        .sector = args->sector
	};

	who = who_info.lookup(&d_k);
	if (who) {
		key.id = who->id;
		info = io_info.lookup_or_try_init(&key, &zero);
		if (info) {
			info->uid = who->uid;
			info->policy = who->policy;
			info->prio = who->prio;
			info->io++;
			info->read_bytes += who->read_bytes;
			info->write_bytes += who->write_bytes;
			__builtin_memcpy(info->comm, who->comm, TASK_COMM_LEN);
			io_info.update(&key, info);
		}

		who_info.delete(&d_k);
	}

   return 0;
}
